class TypeNotBoundError(Exception):
    """Exception raised when a type is not bound to a provider."""


class StateNotInitializedError(Exception):
    """Exception raised when the state is not initialized."""
